/*  $Id$
 *
 *  app.js
 *  Version 1.02 2020
 *  VISD Application
 *                                                       ____________
 *  Copyright (C) 1998-2020 D&D Corporation.             \          /
 *  All rights reserved.                                  \________/
 *  Serż Minus (Sergey Lepenkov), http://www.serzik.com    \      /
 *                                                          \____/
 *  Dual licensed under the MIT and GNU GPL v3. licenses:    \  /
 *   http://www.opensource.org/licenses/mit-license.php       \/
 *   http://www.gnu.org/licenses/gpl.html
 *
*/
"use strict";


/*
    Radio
*/
const channel = Backbone.Radio.channel('VISD');


/*
    Objects
*/
const MyObject = Marionette.MnObject.extend({
    model:      {},
    collection: {},
    view:       {},
    controller: {},
    app:        {},
    timer:      {},
    cntLoading:         0, // For loading bar
    lastdata_is_loaded: false,
    lastdata_counter:   0,
    lastdata_time:      0,
    hostname_main:      undefined,
    hostname:           undefined,
    title:              undefined,
    ahosts:             [],
    ahosts_meta:        {},

    initialize() {
        var self = this;
        //this.search = $.getQSVar('search') || "";
        //this.page = ($.getQSVar('page') || 0) * 1;
        //this.id = ($.getQSVar('id') || 0) * 1;

        // Set timer
        this.timer.minutely = setInterval(function() {
            //self.app.graphsApp.triggerMethod("reload", "Run app trigger reload from MyObject inicialize()");
            self.reload("Run minutelyReload() from MyObject inicialize()");
        }, 20*1000); // Every 1/4 minute
    },
    ajaxErrorHandler(jqXHR, jx, url) {
        jx.status_code = jqXHR.status;
        jx.status_text = jqXHR.statusText;
        var def_msg = "Can't get or parse data content from " + (url || 'server');
        if (jqXHR.responseJSON) {
            jx.message = jqXHR.responseJSON.error || jqXHR.responseJSON.message || def_msg;
        } else {
            jx.status_code = 400;
            jx.status_text = "Can't load JSON document";
            jx.message = jqXHR.responseText || def_msg;
        }
    },
    pushAlert(data) { // Show alert message
        var tpl = _.template($('#tplAlert').html());
        var html = tpl({
            level: data.level || "primary",
            code: data.code || 0,
            title: data.title || "",
            message: data.message || "",
        });
        if (data.noAdd) {
            $('#status-bar').html(html);
        } else {
            $('#status-bar').append(html);
        }
        return true;
    },
    cleanAlert() { // Delete all alerts
        $('#status-bar').empty();
    },
    showLoading() { // Show loading bar
        this.cntLoading++;
        $('#loading').show();
    },
    hideLoading() { // Hide loading bar
        this.cntLoading--;
        if (this.cntLoading <= 0) {
            $('#loading').hide();
        }
    },
    setHostname(hn,t) {
        this.hostname = hn;
        this.setTitle(t || hn);
        this.reload("Change current hostname to: " + hn);
        return this.hostname;
    },
    getHostname() {
        return this.hostname;
    },
    getRootDevice(fs) {
        if (fs === undefined) fs = visd.model.lastdata.get("fs") || {};

        var root = {};
        var catched = [];
        _.each(fs.dev, function(value, key){
            if (value.point == "/") {
                root = {
                    "key": key,
                    total: value.total,
                    used: value.used,
                    id: value.id,
                    point: value.point,
                };
            } else if (! value.opts.match(/nodev/)) {
                catched.push({
                    "key": key,
                    total: value.total,
                    used: value.used,
                    id: value.id,
                    point: value.point,
                });
            }
        });
        if (_.isEmpty(root)) {
            root = _.first(_.sortBy(catched, 'id'));
        }
        if (_.isEmpty(root)) {
            root = undefined;
        }

        return root
    },
    reload(msg) { // Reload all data!
        var self = this;
        var model = this.model.lastdata;
        var url = model.url;

        // Clean before
        model.unset("status");
        model.unset("error");
        model.unset("message");
        model.unset("location");
        model.unset("sys");
        model.unset("fs");
        model.unset("dev");

        // Get data
        var ajx = Object.assign({}, ajaxDefaults);
        var defer = $.Deferred();
        self.showLoading();
        model.fetch({
            timeout: JQX_TIMEOUT,
            cache: false,
            data: {
                'host': self.getHostname(),
            },
            success: function(mdl){
                defer.resolve(mdl);
            },
            error: function(mdl, jqXHR, o){
                self.ajaxErrorHandler(jqXHR, ajx, url);
                defer.resolve(undefined);
            },
        });
        $.when(defer.promise()).done(function(mdl){
            self.hideLoading();
            if (mdl !== undefined) { // OK
                if (self.lastdata_is_loaded == false) {
                    self.lastdata_is_loaded = true;
                    self.cleanAlert();
                }
                // Initialize
                self.lastdata_counter++;
                if (self.lastdata_counter == 1) {
                    self.init_lastdata(mdl);
                } else if (self.lastdata_counter > 1) {
                    self.update_lastdata(mdl);
                }
                channel.trigger("all:reload", msg || "Run channel trigger all:reload from MyObject reload()");
            } else {
                self.lastdata_is_loaded = false;
                self.pushAlert({
                    noAdd: true,
                    level: "danger",
                    code: ajx.status_code,
                    title: ajx.status_text,
                    message: ajx.message,
                });
            }
        });
    },
    init_lastdata(mdl) {
        // Hosts list init
        var hostname = mdl.get("hostname");
        var title = mdl.get("title");
        if (! _.isUndefined(hostname)) {
            // Set main hostname & title
            this.hostname_main = hostname;
            this.title = title || hostname;
            this.setTitle();

            // Set CTime
            this.lastdata_time = mdl.get("ctime");

            // Set main ahosts
            var ahosts = mdl.get("ahosts");
            this.ahosts = ahosts;

            // Set main ahosts_meta
            var ahosts_meta = mdl.get("ahosts_meta");
            this.ahosts_meta = ahosts_meta;

            if (_.isArray(ahosts) && !_.isEmpty(ahosts)) {
                var ahosts_prep = [];
                ahosts_prep.push({
                    "hostname": hostname,
                    "title": title,
                    "is_main": true,
                    "status": statusf(mdl.get("status")),
                });
                _.each(ahosts, function(v){
                    ahosts_prep.push({
                        "hostname": v,
                        "title": ahosts_meta[v].title || v,
                        "is_main": false,
                        "status": statusf( ahosts_meta[v].status ),
                    });
                });
                visd.collection.hosts = new HostCollection(ahosts_prep);
                //console.log(visd.collection.hosts);
            }
        }
    },
    update_lastdata(mdl) {
        var hostname = mdl.get("hostname");

        // check CTime
        var old_time = this.lastdata_time || 0;
        var new_time = mdl.get("ctime") || 0;
        if ((old_time > 0) && (new_time > 0) && (new_time == old_time)) {
            var cur_time = Math.round(new Date().getTime()/1000);
            var delta = cur_time - old_time;
            if (delta > LAST_DATA_EXPIRES) {
                mdl.set("error", "The data is taking too long to update");
                mdl.set("status", -2);
                mdl.set("ctime", new_time + 0.001);
            }
        } else {
            this.lastdata_time = new_time;
        }

        if (visd.hostname_main === hostname) {
            var ahosts = mdl.get("ahosts");
            var ahosts_meta = mdl.get("ahosts_meta");
            if (_.isArray(ahosts)) {
                visd.collection.hosts.each(function(m) {
                    var hn = m.get("hostname");
                    var st;
                    if (hostname === hn) { // Main Hostname
                        st = statusf(mdl.get("status"));
                    } else {
                        st = statusf(ahosts_meta[hn].status);
                    }
                    m.set("status", st);
                });
            }
        }
    },
    setTitle(t) {
        // Change Title & Description
        var title = t || this.title; // hostname_main
        document.title = PROJECT_PREFIX + " · " + title;
        $('meta[name="description"]').attr("content", title);
    },
    run() {
        //console.log("Test message");
        this.reload("Run application");
    },
});
const visd = new MyObject();

/*
    Controllers (Behaviors)
*/
visd.controller.graphboxShow = Marionette.Behavior.extend({ // Link to show of graphbox
    ui: {
        lnkGraphBox: ".lnkGraphBox",
    },
    events: {
        'click @ui.lnkGraphBox': "onClickLnkGraphBox",
    },
    onClickLnkGraphBox(e) {
        e.preventDefault();
        e.stopPropagation();
        this.view.trigger("graphbox:show",
            $(e.target).attr("data-context"),
            $(e.target).attr("data-name")); // Go to next
    },
});
visd.controller.imageReloadable = Marionette.Behavior.extend({
    options: {
        text: 'Reload'
    },
    ui: {
        imageReloadable: '.image-reloadable',
    },
    events: {
        'click @ui.imageReloadable': 'onReloadImage',
        'mouseover @ui.imageReloadable': 'onMouseover',
        'mouseout @ui.imageReloadable': 'onMouseout',
    },
    onReloadImage(event) {
        var src = event.target.src;
        var i = src.indexOf('?_=');
        src = i != -1 ? src.substring(0, i) : src;
        var j = src.indexOf('&_=');
            src = j != -1 ? src.substring(0, j) : src;
        var pfx;
        if (src.indexOf('?') != -1) {
            pfx = '&_='
        } else {
            pfx = '?_='
        }
        var d = new Date();
        event.target.src = src + pfx + d.getTime();
        return false;
    },
    onMouseover(event){
        $(event.target).css("cursor", "pointer");
    },
    onMouseout(event){
        $(event.target).css("cursor", "default");
    },
});
visd.controller.graphChangeZoom = Marionette.Behavior.extend({
    ui: {
        btnChangeZoom: ".btnChangeZoom",
    },
    events: {
        'click @ui.btnChangeZoom': "onClickBtnChangeZoom",
    },
    onClickBtnChangeZoom(e) {
        this.view.triggerMethod("change:zoom", $(e.target).val()); // Go to next
    },
});
visd.controller.tableSort = Marionette.Behavior.extend({
    initialize() {
        this.options.order_field = "";
        this.options.order_direct = 0;
        this.view.sortWithCollection = true;
    },
    ui: {
        th: "th",
    },
    events: {
        'click @ui.th': "onClickTh",
    },
    onClickTh(e) {
        var field = $(e.target).attr("data-order");
        if (_.isEmpty(field)) return false;

        // Get direction
        var direction = 0;
        if (this.options.order_field == field) {
            direction = this.options.order_direct;
            if (direction <= 0) { // Current = ASC
                direction = 1; // ASC
            } else { // DESC
                direction = -1; // DESC
            }
        } else {
            this.options.order_field = field;
            direction = 1; // ASC
        }
        this.options.order_direct = direction;

        
        // Class
        //this.$el.removeClass("active");
        this.getUI("th").removeClass("order-asc order-desc");
        $(e.target).addClass((direction > 0) ? "order-asc" : "order-desc");

        //this.view.triggerMethod("sort:table", field, direction); // Go to view
        this.onSortTable(field, direction);
    },
    onSortTable(field, direct) {
        //console.log("Click to: " + field + " = " + (direct > 0 ? "ASC" : "DESC"));

        //this.view.collection.comparator = field;
        this.view.collection.comparator = function(a, b) {
            var av = a.get(field) || '';
            var bv = b.get(field) || '';
            if (av == bv) return 0;
            if (direct > 0) {
                return (av > bv) ? 1 : -1;
            } else {
                return (av > bv) ? -1 : 1;
            }
        };
        this.view.collection.sort();
    },
});


/*
    Models & Collections
*/
const LastdataModel = Backbone.Model.extend({
    url: JQX_PFX_URL,
    defaults: {
        "status": -2, // unknown
        "message" : [],
        "error": undefined,
        "hostname": "localhost",
        "title": "",
        "location": "",
        "iter": 0,
        "uptime": 0,
        "ctime": 0,
        "cpuusage": 0,
        "processes": 0,
        "idle": 0,
        "iodelay": 0,
        "iowait": 0,
        "kversion": "",
        "kmachine": "",
        "version": "",
        "banner": "",
        "ahosts": [],
        "ahosts_meta": {},
        "loadavg": [0,0,0],
        "cpuinfo": {
            "user_hz": 0,
            "mhz": 0,
            "sockets": 0,
            "hvm": 0,
            "model": "",
            "cpus": 0,
        },
        "memory" : {
            "shared" : 0,
            "free" : 0,
            "used" : 0,
            "buffer" : 0,
            "total" : 0,
        },
        "swap": {
            "free": 0,
            "used": 0,
            "total": 0,
        },
        "net": {},
        "fs": {},
        "sys": {},
    },
});
visd.model.lastdata = new LastdataModel();

const HostModel = Backbone.Model.extend({
    defaults: {
        "hostname": "",
        "title": "",
        "is_main": false,
        "status": statusf(), // unknown
    },
});
const HostCollection = Backbone.Collection.extend({
    model: HostModel,
});
visd.collection.hosts = new HostCollection();

const GraphModel = Backbone.Model.extend({
    defaults: {
        "rownum": 0,
        "title": "",
        "host": "",
        "group": "sys",
        "name": "",
    },
});
const GraphCollection = Backbone.Collection.extend({
    model: GraphModel,
});

const DeviceModel = Backbone.Model.extend({
    defaults: {
        devid: 0,
        majmin: "",
        device: "",
        point: "",
        fstype: "",
        opts: "",
        total: 0,
        avail: 0,
        free: 0,
        used: 0,
    },
});
const DeviceCollection = Backbone.Collection.extend({
    comparator: "devid",
    model: DeviceModel,
});

const InterfaceModel = Backbone.Model.extend({
    defaults: {
        imterface: "",
        transmit: 0,
        receive: 0,
    },
});
const InterfaceCollection = Backbone.Collection.extend({
    comparator: "interface",
    model: InterfaceModel,
});

const RouteIPv4Model = Backbone.Model.extend({
    defaults: {
        destination: "",
        gateway: "",
        iface: "",
        mask: "",
        metric: 0,
        mtu: "",
    },
});
const RouteIPv6Model = Backbone.Model.extend({
    defaults: {
        destination: "",
        gateway: "",
        iface: "",
        metric: 0,
        prefix: 0,
    },
});
const RouteIPv4Collection = Backbone.Collection.extend({
    model: RouteIPv4Model,
});
const RouteIPv6Collection = Backbone.Collection.extend({
    model: RouteIPv6Model,
});

const SocketModel = Backbone.Model.extend({
    defaults: {
        inode: "",
        sl: "",
        proto: "",
        local_ip : "",
        local_port : "",
        rem_ip: "",
        rem_port: "",
        st: "",
        uid: 0,
        user: "",
    },
});
const SocketsCollection = Backbone.Collection.extend({
    model: SocketModel,
});

const ProcessModel = Backbone.Model.extend({
    defaults: {
        size: "",
        state: "",
        pctmem: "",
        fname: "",
        start: "",
        rss: "",
        user: "",
        cmdline: "",
        uid: 0,
        ttydev: "",
        ppid: 0,
        pid: 0,
        pctcpu: "",
    },
});
const ProcessesCollection = Backbone.Collection.extend({
    model: ProcessModel,
});


/*
    Views
*/
visd.view.empty = Marionette.View.extend({ // visd.view.empty({message: msg});
    tagName: "span",
    className: "text-muted p-3",
    template: _.template('<%= message %>'),
    initialize: function(options){
        var opts = options || {};
        this.message = opts.message || "No data";
    },
    serializeData: function(){
        return {
            message: this.message,
        };
    },
});
visd.view.emptyDiv = Marionette.View.extend({ // visd.view.emptyDiv({message: msg});
    tagName: "div",
    template: _.template('<p><%= message %></p>'),
    initialize: function(opts){
        this.options.message = opts.message || "No data";
    },
    serializeData: function(){
        return {
            message: this.options.message,
        };
    },
});
visd.view.alert = Marionette.View.extend({
    template: _.template($('#tplAlert').html()),
    initialize: function(options){
        var opts = options || {};
        this.level  = opts.level || "danger"; // primary, secondary, success, danger, warning, info, light, dark  # https://getbootstrap.com/docs/4.1/components/alerts/ 
        this.code   = opts.code || 0;
        this.title  = opts.title || "Error!";
        this.message= opts.message || "unknown error";
    },
    serializeData: function(){
        var code = this.code;
        var msg  = this.message;
        if (!code && msg == "timeout") {
            code = "-408";
            msg = "Timed out!";
        } else if (!code && msg == "error") {
            code = "-400";
            msg = "Can't connect to server!";
        }
        return {
            level: this.level,
            title: this.title,
            code: code,
            message: msg,
        };
    },
});
visd.view.graphbox = Marionette.View.extend({
    template: _.template($('#tplGraphBox').html()),
    ui: {
        modalGraphBox: "#modalGraphBox",
        imgGraphBox: "#imgGraphBox",
        btnShowGraphs: "#btnShowGraphs",
    },
    behaviors: {
        imageReloadable: visd.controller.imageReloadable,
        graphChangeZoom: visd.controller.graphChangeZoom,
    },
    events: {
        'click @ui.btnShowGraphs': 'onShowGraphs',
    },
    onShowGraphs(e) {
        e.preventDefault();
        e.stopPropagation();
        this.ui.modalGraphBox.modal('hide');
        $("#graphs-tab").tab('show');
    },
    onChangeZoom(zoom) {
        var host = visd.getHostname() || visd.hostname_main;
        this.ui.imgGraphBox.attr("src", this.mkURL(host, this.options.group, this.options.name, zoom));
    },
    templateContext() {
        var graph = this.options.graph || "sys";
        var host = visd.getHostname() || visd.hostname_main;

        var url; //this.mkURL(host, group, name, zoom)
        var title = "No title";

        if (graph == "sys") {
            title = "System graph for " + host;
            this.options.group = "sys";
        } else if (graph == "load") {
            title = "System load graph for " + host;
            this.options.group = "load";
        } else if (graph == "root_usage") {
            var root = visd.getRootDevice(this.model.get("fs")) || {};
            var key = root.key || "unknown";
            title = "Disk (" + key + ") usage graph for " + host;
            this.options.group = "dev_usage";
            this.options.name = key;
        } else if (graph == "root_io") {
            var root = visd.getRootDevice(this.model.get("fs")) || {};
            var key = root.key || "unknown";
            title = "Disk (" + key + ") IO graph for " + host;
            this.options.group = "dev_io";
            this.options.name = key;
        } else if (graph == "dev_usage") {
            title = "Disk (" + this.options.name + ") usage graph for " + host;
            this.options.group = "dev_usage";
        } else if (graph == "dev_io") {
            title = "Disk (" + this.options.name + ") IO graph for " + host;
            this.options.group = "dev_io";
        } else if (graph == "net") {
            title = "Network interface (" + this.options.name + ") graph for " + host;
            this.options.group = "net";
        }

        return {
            "title": this.options.title || title,
            "graph": graph,
            "url": this.mkURL(host, this.options.group, this.options.name, this.options.zoom),
        };
    },
    showModal(data) {
        this.options.title = data.title;
        this.options.graph = data.graph;
        this.options.name = data.name;
        this.options.zoom = data.zoom;
        this.render();
        this.ui.modalGraphBox.modal('show');
    },
    mkURL(host, group, name, zoom) {
        var url = JQX_PFX_URL + "?action=graph";
        if (host) url = url + "&host=" + host;
        if (group) url = url + "&group=" + group;
        if (name) url = url + "&name=" + name;
        if (zoom) url = url + "&zoom=" + zoom;
        var d = new Date();
        return url + "&_=" + d.getTime();
    },
});

// NavBar (hosts)
visd.view.hostitem = Marionette.View.extend({
    tagName: "a", // <a class="nav-item nav-link active" href="#" data-toggle="pill" role="tab">Active</a>
    className: "nav-item nav-link",
    attributes: {
        "data-toggle": "pill",
        "href": "#",
        "role": "tab",
    },
    template: _.template('<span class="<%= iclass %>" style="<%= istyle %>"></span><span class="<%= aclass %>"><%= title %></span>'),
    onRender() {
        // Remove classes "active"
        this.$el.removeClass("active");

        // Add classes "active"
        var cur_hostname = visd.getHostname("hostname");
        var hostname = this.model.get("hostname");
        if (_.isUndefined(cur_hostname)) {
            if (this.model.get("is_main")) this.$el.addClass("active");
        } else {
            if (hostname == cur_hostname) this.$el.addClass("active");
        }

        // Status
        var status = this.model.get("status");
        this.$el.attr("title", "Status of " + hostname + ": " + status);
    },
    templateContext() {
        var status = this.model.get("status");
        // Status class
        if (status === "good") {
            return {
                "iclass": "far fa-check-circle",
                "istyle": "color: #21bf4b !important",
                "aclass": "ml-1",
            };
        } else if (status === "warning") {
            return {
                "iclass": "fas fa-exclamation-circle",
                "istyle": "color: #ffcc00 !important",
                "aclass": "status_warning font-weight-bold ml-1",
            };
        } else if (status === "critical") {
            return {
                "iclass": "fas fa-times-circle",
                "istyle": "color: #ff6c59 !important",
                "aclass": "status_critical font-weight-bold ml-1",
            };
        } else { // unknown
            return {
                "iclass": "far fa-question-circle",
                "istyle": "color: #cfcfcf !important",
                "aclass": "status_unknown ml-1",
            };
        }
    },
    modelEvents: {
        'change:status': 'onChangeModelStatus',
    },
    onChangeModelStatus(model, value) {
        this.render();
    },
    events: {
        click: 'onClickItem',
    },
    onClickItem(e) {
        e.preventDefault();
        //console.log("Pressed: " + this.$el.text());
        //$(this).tab('show')
        this.trigger("hostname:change", this); // Go to next
        return true;
    },
});
visd.view.hostsnavbar = Marionette.CollectionView.extend({
    tagName: "nav",
    className: "nav nav-pills flex-column flex-sm-row mb-4",
    attributes: {
        "role": "tablist",
        "aria-label": "Host select",
    },
    childView: visd.view.hostitem,
    childViewEvents: {
        'hostname:change': 'onHostnameChange',
    },
    onHostnameChange(v) {
        this.trigger("child:hostname:change", v.model.get("hostname"), v.model.get("title")); // Go to next
    },
});

// Section: Health
visd.view.sectionHealth = Marionette.View.extend({
    template: _.template($('#tplSectionHealth').html()),
    model: visd.model.lastdata,
    behaviors: {
        graphboxShow: visd.controller.graphboxShow,
    },
    templateContext() {
        var cpuinfo = this.model.get("cpuinfo") || {};
        var cpus = (cpuinfo.cpus || 1) * (cpuinfo.sockets || 1);

        // CPU usage
        var cpuusage = (this.model.get("cpuusage") || 0) * 100;
        var cpuusagef = Math.round((cpuusage) * 100)/100;

        // IO delay
        var iodelay = (this.model.get("iodelay") || 0) * 100;
        var iodelayf = Math.round((iodelay) * 100)/100;

        // RAM
        var ram = this.model.get("memory") || {};
        var ram_used = ram.used || 0;
        var ram_total = ram.total || 0;

        // Swap
        var swp = this.model.get("swap") || {};
        var swp_used = swp.used || 0;
        var swp_total = swp.total || 0;

        // Root
        var root = visd.getRootDevice(this.model.get("fs")) || {};
        var root_used = root.used || 0;
        var root_total = root.total || 0;

        // Time
        var srv_ctime = this.model.get("ctime") || 0;
        var srv_uptime = this.model.get("uptime") || 0;
        var srv_started = srv_ctime - srv_uptime;

        return {
            "statusf": statusf(this.model.get("status")),
            "statusc": status_color(this.model.get("status")),
            "ctimef": timeConverter(srv_ctime),
            "uptimef": fduration(srv_uptime),
            "started": srv_started > 0 ? timeConverter(srv_started) : "Unknown",
            "cpuusagef": cpuusagef,
            "cpus": cpus,
            "iodelayf": iodelayf,
            "location": this.model.get("location") || "Unknown",
            "message": this.model.get("error") || this.model.get("message") || "No messages",

            // RAM
            "ram_usage": ram_total ? Math.round(100 * 100 * ram_used / ram_total)/100 : 0,
            "ram_used": formatBytes(ram_used),
            "ram_total": formatBytes(ram_total),
            "ram_buffer": formatBytes(ram.buffer || 0),
            "ram_shared": formatBytes(ram.shared || 0),

            // Swap
            "swp_usage": swp_total ? Math.round(100 * 100 * swp_used / swp_total)/100 : 0,
            "swp_used": formatBytes(swp_used),
            "swp_total": formatBytes(swp_total),

            // Root
            "root_usage": root_total ? Math.round(100 * 100 * root_used / root_total)/100 : 0,
            "root_used": formatBytes(root_used),
            "root_total": formatBytes(root_total),
            "root_key": root.key || "unknown",
            "root_point": root.point || "root",
        };
    },
    modelEvents: {
        'change:ctime': 'onChangeModelCTime',
    },
    onChangeModelCTime(model, value) {
        this.render();
    },
});

// Section: Graphs
visd.view.graph = Marionette.View.extend({
    tagName: "div",
    className: "mb-3",
    template: _.template($('#tplGraphsRow').html()),
    behaviors: {
        imageReloadable: visd.controller.imageReloadable,
        graphChangeZoom: visd.controller.graphChangeZoom,
    },
    ui: {
        imgGraph: "img.image-graph",
    },
    onChangeZoom(zoom) {
        this.ui.imgGraph[0].src = this.mkURL(zoom);
    },
    mkURL(zoom) {
        var url = JQX_PFX_URL + "?action=graph";
        if (this.model.get("host")) {
            url = url + "&host=" + this.model.get("host");
        }
        if (this.model.get("group")) {
            url = url + "&group=" + this.model.get("group");
        }
        if (this.model.get("name")) {
            url = url + "&name=" + this.model.get("name");
        }
        if (zoom) {
            url = url + "&zoom=" + zoom;
        }
        var d = new Date();
        return url + "&_=" + d.getTime();
    },
    templateContext() {
        return {
            "url": this.mkURL(),
        };
    },
});
visd.view.graphs = Marionette.CollectionView.extend({
    tagName: "div",
    childView: visd.view.graph,
});
visd.view.sectionGraphs = Marionette.View.extend({
    template: _.template($('#tplSectionGraphs').html()),
    regions: {
        regionGraphs: "#regionGraphs",
    },
    model: visd.model.lastdata,
    modelEvents: {
        'change:hostname': 'onChangeModelHostname',
    },
    onChangeModelHostname(model, value) {
        this.render();
    },
    onRender() {
        var graphs = [];
        var host = this.model.get("hostname") || "localhost";
        var i = 0;

        // Add system
        graphs.push({
            rownum: ++i,
            title: "System for " + host,
            host: host,
            group: "sys",
            name: undefined,
        });

        // Add system load
        graphs.push({
            rownum: ++i,
            title: "System load for " + host,
            host: host,
            group: "load",
            name: undefined,
        });

        // Add devices
        var fs = this.model.get("fs") || {};
        _.each(fs.dev, function(value, key){
            graphs.push({
                rownum: ++i,
                title: "Disk (" + key + ") usage for " + host,
                host: host,
                group: "dev_usage",
                name: key,
            });
            graphs.push({
                rownum: ++i,
                title: "Disk (" + key + ") IO for " + host,
                host: host,
                group: "dev_io",
                name: key,
            });
        });

        // Add interfaces
        var net = this.model.get("net");
        _.each(net.interfaces, function(value, key){
            graphs.push({
                rownum: ++i,
                title: "Interface (" + key + ") for " + host,
                host: host,
                group: "net",
                name: key,
            });
        });

        // Render graph list
        this.showChildView('regionGraphs', new visd.view.graphs({
            collection: new GraphCollection(graphs),
        }));
    },
});

// Section: Devices
visd.view.device = Marionette.View.extend({
    tagName: "div",
    className: "mb-3",
    template: _.template($('#tplDevicesRow').html()),
    behaviors: {
        graphboxShow: visd.controller.graphboxShow,
    },
    templateContext() {
        var used = this.model.get("used") || 0;
        var total = this.model.get("total") || 0;
        var opts = this.model.get("opts") || "";
        return {
            "usage": total ? Math.round(100 * 100 * used / total)/100 : 0,
            "usedf": formatBytes(used),
            "totalf": formatBytes(total),
            "availf": formatBytes(this.model.get("avail") || 0),
            "freef": formatBytes(this.model.get("free") || 0),
            "isDevice": opts.match(/nodev/i) ? 0 : 1,
        };
    },
});
visd.view.devices = Marionette.CollectionView.extend({
    tagName: "div",
    childView: visd.view.device,
    childViewTriggers: {  // Proxy to visd.view.sectionDevices
        "graphbox:show": "child:graphbox:show",
    },
    emptyView: visd.view.emptyDiv,
    emptyViewOptions: {
        message: "No any device available. Please check VISD configuration file!",
        className: "text-danger font-weight-bold"
    },
});
visd.view.sectionDevices = Marionette.View.extend({
    template: _.template($('#tplSectionDevices').html()),
    regions: {
        regionDevices: "#regionDevices",
    },
    model: visd.model.lastdata,
    modelEvents: {
        'change:ctime': 'onChangeModelCTime',
    },
    onChangeModelCTime(model, value) {
        this.render();
    },
    childViewTriggers: { // Proxy to visd.view.lastdata
        "child:graphbox:show": "graphbox:show",
    },
    onRender() {
        var devices = [];

        var fs = this.model.get("fs");
        _.each(fs.dev, function(value, key){
            devices.push({
                devid: value.id || 0,
                majmin: value.majmin || "",
                device: key,
                point: value.point || "",
                fstype: value.fstype || "",
                opts: value.opts || "",
                // Sizes
                total: value.total,
                avail: value.avail,
                free: value.free,
                used: value.used,
            });
        });

        // Render device list
        this.showChildView('regionDevices', new visd.view.devices({
           collection: new DeviceCollection(devices),
        }));
    },
});

// Section: Interfaces
visd.view.interface = Marionette.View.extend({
    tagName: "div",
    className: "mb-3",
    template: _.template($('#tplInterfacesRow').html()),
    behaviors: {
        graphboxShow: visd.controller.graphboxShow,
    },
    templateContext() {
        var transmit = this.model.get("transmit") || 0;
        var receive = this.model.get("receive") || 0;
        return {
            "transmitf": formatBytes(transmit),
            "transmitfd": digsf(transmit),
            "receivef": formatBytes(receive),
            "receivefd": digsf(receive),
        };
    },
});
visd.view.interfaces = Marionette.CollectionView.extend({
    tagName: "div",
    childView: visd.view.interface,
    childViewTriggers: {  // Proxy to visd.view.sectionInterfaces
        "graphbox:show": "child:graphbox:show",
    },
    emptyView: visd.view.emptyDiv,
    emptyViewOptions: {
        message: "No any network interface available. Please check VISD configuration file!",
        className: "text-danger font-weight-bold"
    },
});
visd.view.sectionInterfaces = Marionette.View.extend({
    template: _.template($('#tplSectionInterfaces').html()),
    regions: {
        regionInterfaces: "#regionInterfaces",
    },
    model: visd.model.lastdata,
    modelEvents: {
        'change:ctime': 'onChangeModelCTime',
    },
    onChangeModelCTime(model, value) {
        this.render();
    },
    childViewTriggers: { // Proxy to visd.view.lastdata
        "child:graphbox:show": "graphbox:show",
    },
    onRender() {
        var interfaces = [];

        var net = this.model.get("net");
        _.each(net.interfaces, function(value, key){
            interfaces.push({
                interface: key,
                transmit: value.transmit || 0,
                receive: value.receive || 0,
            });
        });

        // Render interface list
        this.showChildView('regionInterfaces', new visd.view.interfaces({
           collection: new InterfaceCollection(interfaces),
        }));
    },
});

// Section: Routes
visd.view.routeIPv4 = Marionette.View.extend({
    tagName: "tr",
    className: "mb-3",
    template: _.template($('#tplRoutesIPv4Row').html()),
});
visd.view.routeIPv6 = Marionette.View.extend({
    tagName: "tr",
    className: "mb-3",
    template: _.template($('#tplRoutesIPv6Row').html()),
    templateContext() {
        return {
            "destinationf": compressIPV6(this.model.get("destination") || 0),
            "gatewayf": compressIPV6(this.model.get("gateway") || 0),
        };
    },
});
visd.view.routesIPv4 = Marionette.CollectionView.extend({
    tagName: "table",
    className: "table table-sm table-bordered table-hover table-responsive-sm",
    template: _.template($('#tplRoutesIPv4Header').html()),
    childViewContainer: 'tbody',
    childView: visd.view.routeIPv4,
    emptyView: visd.view.emptyDiv,
    emptyViewOptions: {
        message: "No any IPv4 route",
        className: "text-secondary"
    },
    behaviors: {
        tableSort: visd.controller.tableSort,
    },
});
visd.view.routesIPv6 = Marionette.CollectionView.extend({
    tagName: "table",
    className: "table table-sm table-bordered table-hover table-responsive-sm",
    template: _.template($('#tplRoutesIPv6Header').html()),
    childViewContainer: 'tbody',
    childView: visd.view.routeIPv6,
    emptyView: visd.view.emptyDiv,
    emptyViewOptions: {
        message: "No any IPv6 route",
        className: "text-secondary"
    },
    behaviors: {
        tableSort: visd.controller.tableSort,
    },
});
visd.view.sectionRoutes = Marionette.View.extend({
    template: _.template($('#tplSectionRoutes').html()),
    regions: {
        regionRoutesIPv4: "#regionRoutesIPv4",
        regionRoutesIPv6: "#regionRoutesIPv6",
    },
    model: visd.model.lastdata,
    modelEvents: {
        'change:hostname': 'onChangeModelHostname',
    },
    onChangeModelHostname(model, value) {
        this.render();
    },
    onRender() {
        var net = this.model.get("net");
        var ipv4 = net.route.ipv4;
        if (! _.isArray(ipv4)) ipv4 = [];
        var ipv6 = net.route.ipv6;
        if (! _.isArray(ipv6)) ipv6 = [];

        // Render route list
        this.showChildView('regionRoutesIPv4', new visd.view.routesIPv4({
           collection: new RouteIPv4Collection(ipv4),
        }));
        this.showChildView('regionRoutesIPv6', new visd.view.routesIPv6({
           collection: new RouteIPv6Collection(ipv6),
        }));
    },
});

// Section: Sockets
visd.view.socket = Marionette.View.extend({
    tagName: "tr",
    className: "mb-3",
    template: _.template($('#tplSocketsRow').html()),
});
visd.view.sockets = Marionette.CollectionView.extend({
    tagName: "table",
    className: "table table-sm table-bordered table-hover table-responsive-sm",
    template: _.template($('#tplSocketsHeader').html()),
    childViewContainer: 'tbody',
    childView: visd.view.socket,
    emptyView: visd.view.emptyDiv,
    emptyViewOptions: {
        message: "No any socket found",
        className: "text-secondary"
    },
    behaviors: {
        tableSort: visd.controller.tableSort,
    },
});
visd.view.sectionSockets = Marionette.View.extend({
    template: _.template($('#tplSectionSockets').html()),
    regions: {
        regionSocketsTCP: "#regionSocketsTCP",
        regionSocketsUDP: "#regionSocketsUDP",
    },
    model: visd.model.lastdata,
    modelEvents: {
        'change:hostname': 'onChangeModelHostname',
    },
    onChangeModelHostname(model, value) {
        this.render();
    },
    onRender() {
        var net = this.model.get("net");
        var sockets = net.sockets;
        var tcp = [];
        var udp = [];
        if (sockets !== undefined) {
            tcp = sockets.tcp;
            if (! _.isArray(tcp)) tcp = [];
            udp = sockets.udp;
            if (! _.isArray(udp)) udp = [];
        }

        // Render lists
        this.showChildView('regionSocketsTCP', new visd.view.sockets({
           collection: new SocketsCollection(tcp),
        }));
        this.showChildView('regionSocketsUDP', new visd.view.sockets({
           collection: new SocketsCollection(udp),
        }));
    },
});

// Section: Processes
visd.view.process = Marionette.View.extend({
    tagName: "tr",
    className: "mb-3",
    template: _.template($('#tplProcessesRow').html()),
    templateContext() {
        var t = this.model.get("start") || 0;
        var cur = Math.round(_.now()/1000);
        return {
            "startf": t ? dtf_short(t) : "?",
            "timef": t ? fduration(cur - t) : "?",
            "command": _.escape(this.model.get("cmndline") || ""),
            "sizef": formatBytes(this.model.get("size") || 0),
            "rssf": formatBytes(this.model.get("rss") || 0),
        };
    },
    ui: {
        tooltip: "td.mytooltip",
    },
    onRender() {
        $(this.ui.tooltip).tooltip();
    },
});
visd.view.processes = Marionette.CollectionView.extend({
    tagName: "table",
    className: "table table-sm table-bordered table-hover table-responsive-sm",
    template: _.template($('#tplProcessesHeader').html()),
    childViewContainer: 'tbody',
    childView: visd.view.process,
    emptyView: visd.view.emptyDiv,
    emptyViewOptions: {
        message: "No any process found",
        className: "text-secondary"
    },
    behaviors: {
        tableSort: visd.controller.tableSort,
    },
});
visd.view.sectionProcesses = Marionette.View.extend({
    template: _.template($('#tplSectionProcesses').html()),
    regions: {
        regionProcesses: "#regionProcesses",
    },
    model: visd.model.lastdata,
    initialize() {
        this.model.set("procs", 0);
        this.processes = [];
    },
    modelEvents: {
        'change:hostname': 'onChangeModelHostname',
    },
    onChangeModelHostname(model, value) {
        this.render();
    },
    onBeforeRender() {
        var sys = this.model.get("sys");
        var processes = [];
        if (sys !== undefined) {
            processes = sys.processes;
        }
        if (! _.isArray(processes)) {
            processes = [];
        }
        this.model.set("procs", _.size(processes));
        this.processes = processes;
    },
    onRender() {
        // Render list
        this.showChildView('regionProcesses', new visd.view.processes({
           collection: new ProcessesCollection(this.processes),
        }));
    },
});



// Application Root Views
visd.view.lastdata = Marionette.View.extend({
    tagName:  "div",
    template: _.template($('#tplLastDataApp').html()),
    regions: {
        regionHostsNavbar: "#regionHostsNavbar",
        regionGraphBox: "#regionGraphBox",
        regionSectionHealth: "#health",
        regionSectionGraphs: "#graphs",
        regionSectionInterfaces: "#interfaces",
        regionSectionDevices: "#devices",
        regionSectionRoutes: "#routes",
        regionSectionSockets: "#sockets",
        regionSectionProcesses: "#processes",
    },
    initialize(options) {
        var data = options || {};
        this.options.is_initialized = data.is_initialized || false;
    },
    renderResult() {
        var self = this;

        // Render Section Health
        this.showChildView('regionSectionHealth', new visd.view.sectionHealth());

        // Render Section Graphs
        this.showChildView('regionSectionGraphs', new visd.view.sectionGraphs());

        // Render Section Interfaces
        this.showChildView('regionSectionInterfaces', new visd.view.sectionInterfaces());

        // Render Section Routes
        this.showChildView('regionSectionRoutes', new visd.view.sectionRoutes());

        // Render Section Devices
        this.showChildView('regionSectionDevices', new visd.view.sectionDevices());

        // Render Section Sockets
        this.showChildView('regionSectionSockets', new visd.view.sectionSockets());

        // Render Section Processes
        this.showChildView('regionSectionProcesses', new visd.view.sectionProcesses());

        // Render GraphBox
        this.showChildView('regionGraphBox', new visd.view.graphbox({
            model: visd.model.lastdata,
        }));
    },
    onLastdataReload(msg) {
        if (this.options.is_initialized == false) { // First call only
            // Render all! (first call only)
            this.render();

            // Render hosts navbar
            this.showChildView('regionHostsNavbar', new visd.view.hostsnavbar({
                collection: visd.collection.hosts,
            }));

            this.options.is_initialized = true;
            this.renderResult();
        }
        //console.log("visd.view.lastdata: onLastdataReload! " + msg || "");
        //this.showChildView('mainRegion', new mdscore.view.loading());
    },
    onGraphsReload(msg) {
        //console.log(msg);
        this.getChildView('regionSectionGraphs').render();
        
    },
    childViewEvents: {
        'child:hostname:change': 'onHostnameChange',
        'graphbox:show': 'onGraphboxShow'
    },
    onHostnameChange(hostname, title) {
        //console.log("Pressed: " + hostname);
        visd.setHostname(hostname, title);
    },
    onGraphboxShow(graph, name) {
        //this.trigger("child:hostname:change", v.model.get("hostname")); // Go to next
        this.getChildView('regionGraphBox').showModal({
            graph: graph,
            name: name,
        });
    },
});
visd.view.about = Marionette.View.extend({
    tagName:  "div",
    template: _.template($('#tplAboutApp').html()),
    model: visd.model.lastdata,
    renderResult() {
        var self = this;
        this.render();
    },
    onLastdataReload(msg) {
        //console.log("visd.view.about: onLastdataReload! " + msg || "");
        //this.showChildView('mainRegion', new mdscore.view.loading());
        this.renderResult();
    },
});


/*
    Applications
*/
const LastDataApp = Marionette.Application.extend({ // LastData
    region: "#LastDataApp", // #LastDataApp
    onStart() {
        this.lastdataView = new visd.view.lastdata();
        this.showView(this.lastdataView);
    },
    initialize() {
        this.listenTo(channel, 'all:reload', this.onReload);
        this.listenTo(channel, 'graphs:reload', this.onGraphsReload);
    },
    onReload(msg) {
        this.lastdataView.triggerMethod('lastdata:reload', msg);
    },
    onGraphsReload(msg) {
        this.lastdataView.triggerMethod('graphs:reload', msg);
    },
    getModel() {
        return this.lastdataView.model;
    },
});
visd.app.lastdata = new LastDataApp();
visd.app.lastdata.start();

const AboutApp = Marionette.Application.extend({ // AboutApp
    region: "#AboutApp", // #AboutApp
    onStart() {
        this.aboutView = new visd.view.about();
        this.showView(this.aboutView);
    },
    initialize() {
        this.options.isInitialized = false;
        this.listenTo(channel, 'all:reload', this.onReload);
    },
    onReload(msg) {
        if (this.options.isInitialized) return true;
        if (visd.model.lastdata.get("version")) {
            // Render about databox
            this.aboutView.triggerMethod('lastdata:reload', msg);
            this.options.isInitialized = true;
        }
    },
});
visd.app.about = new AboutApp();
visd.app.about.start();

visd.run();

/* __END__ */